/**
 * PETALS - PETALS Services Platform. Copyright (c) 2008 OW2 consortium,
 * http://www.ow2.org/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.binding.soap.listener.outgoing;

import java.net.URI;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.messaging.MessagingException;
import javax.xml.namespace.QName;

import org.apache.axis2.AxisFault;
import org.apache.axis2.Constants;
import org.apache.axis2.addressing.EndpointReference;
import org.apache.axis2.client.Options;
import org.apache.axis2.client.ServiceClient;
import org.apache.axis2.context.ConfigurationContext;
import org.apache.axis2.context.ConfigurationContextFactory;
import org.apache.axis2.description.AxisOperation;
import org.apache.axis2.description.AxisService;
import org.apache.axis2.description.OutInAxisOperation;
import org.apache.axis2.description.OutOnlyAxisOperation;
import org.apache.axis2.description.RobustOutOnlyAxisOperation;
import org.apache.axis2.transport.http.HTTPConstants;
import org.apache.axis2.transport.http.HttpTransportProperties;
import org.apache.commons.pool.BasePoolableObjectFactory;
import org.ow2.petals.binding.soap.listener.SUPropertiesHelper;
import org.ow2.petals.binding.soap.listener.incoming.PetalsAxisService;
import org.ow2.petals.component.framework.Constants.MEPConstants;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.util.IDGenerator;

import static org.ow2.petals.binding.soap.Constants.ServiceUnit.ADDRESS;

public class ServiceClientPoolObjectFactory extends BasePoolableObjectFactory {

    private final QName operation;

    private final URI mep;

    private final ConfigurationExtensions cdkExtensions;
    
    private final List<String> modules;
    
    private final Logger logger;
    
    /**
     * The configuration context. Reuse always the same in all the
     * {@link ServiceClient}s.
     */
    private static ConfigurationContext axisCtx = null;

    static {
        try {
            axisCtx = ConfigurationContextFactory.createConfigurationContextFromFileSystem(null,
                    null);
        } catch (final AxisFault ae) {
            System.err.println("WARNING : Error while creating the Axis Configuration Context");
        }
    }

    public ServiceClientPoolObjectFactory(final QName operation, final URI mep,
            final ConfigurationExtensions cdkExtensions, final List<String> modules, final Logger logger) {
        this.operation = operation;
        this.mep = mep;
        this.cdkExtensions = cdkExtensions;
        this.modules = modules;
        this.logger = logger;
    }

    @Override
    public Object makeObject() throws Exception {

        if (this.logger.isLoggable(Level.FINE)) {
            this.logger.fine("Creating a service client for " + this.cdkExtensions.get(ADDRESS)
                    + ", " + this.operation + ", " + this.mep);
        }

        final AxisService service = new AxisService(
                PetalsAxisService.OUTGOING_SERVICE_CLIENT_PREFIX
                        + IDGenerator.getInstance().getNewID());

        final AxisOperation axisOperation;
        if (MEPConstants.IN_ONLY_PATTERN.equals(mep)) {
            axisOperation = new OutOnlyAxisOperation(operation);
        } else if (MEPConstants.ROBUST_IN_ONLY_PATTERN.equals(mep)) {
            axisOperation = new RobustOutOnlyAxisOperation(operation);
        } else if (MEPConstants.IN_OPTIONAL_OUT_PATTERN.equals(mep)
                || MEPConstants.IN_OUT_PATTERN.equals(mep)) {
            axisOperation = new OutInAxisOperation(operation);
        } else {
            axisOperation = null;
        }
        service.addOperation(axisOperation);

        // create Options for the stub
        final Options options = ServiceClientPoolObjectFactory.createOptions(operation, cdkExtensions);

        try {
            final ServiceClient serviceClient = new ServiceClient(axisCtx, service);
            serviceClient.setOptions(options);
            
            // engage the Axis2 modules
            this.engageModules(serviceClient);
            
            // TODO: Implement enableSSLSupport
            this.enableSSLSupport(this.cdkExtensions);
            
            return serviceClient;
        } catch (final AxisFault e) {
            throw new MessagingException("Can't create ServiceClient", e);
        }
    }
    
    /**
     * Create the SOAP options of the outgoing SOAP message
     * 
     * @param address
     * @param operation
     * @param extensions
     * @return
     */
    private static Options createOptions(final QName soapAction,
            final ConfigurationExtensions extensions) {

        final Options options = new Options();
        
        final String mode = SUPropertiesHelper.getMode(extensions);
        if ("SOAP".equals(mode)) {

            // set destination address
            options.setTo(new EndpointReference(extensions.get(ADDRESS)));
    
            // set the soap action
            if (soapAction != null) {
                options.setAction(soapAction.toString());
            }
    
            // set the timeout
            final long timeout = SUPropertiesHelper.retrieveTimeout(extensions);
            if (timeout != -1L) {
                options.setTimeOutInMilliSeconds(timeout);
            }
    
            // Get the soapEnvelopeNamespaceURI version to use - optional, default
            // is 1.1 version
            final String soapEnvelopeNamespaceURI = AbstractJBIDispatcher.retrieveSOAPEnvelopeNamespaceURI(extensions);
            options.setSoapVersionURI(soapEnvelopeNamespaceURI);
    
            // set transport
            options.setTransportInProtocol(Constants.TRANSPORT_HTTP);
    
            // get proxy settings if they are defined in the extensions
            final HttpTransportProperties.ProxyProperties proxyProperties = AbstractJBIDispatcher
                    .retrieveProxySettings(extensions);
            if (proxyProperties != null) {
                options.setProperty(HTTPConstants.PROXY, proxyProperties);
            }
    
            options.setProperty(org.apache.axis2.transport.http.HTTPConstants.CHUNKED, AbstractJBIDispatcher
                    .retrieveChunkedMode(extensions));
    
            options.setCallTransportCleanup(AbstractJBIDispatcher.retrieveCleanupTransport(extensions));
        }
        else if ("REST".equals(mode)){
            
            // TODO: Move here options from RESTCaller.java when optimizing REST calls using the pool

        }
        return options;
    }

    /**
     * Engage the modules for the service client. The modules are available on
     * the service unit listener since they have been defined during SU
     * deployment.
     * 
     * @param serviceClient
     * @throws AxisFault
     */
    private void engageModules(final ServiceClient serviceClient)
            throws AxisFault {
        
        if (this.modules != null) {
            for (final String name : modules) {
                this.logger.fine("Engaging module " + name);
                serviceClient.engageModule(name);
            }
        }
    }

    /**
     * TODO Enable the client to contact a service over an SSL connection. The
     * required values are in the PEtALS extensions if the SU has been
     * parametrized to enable SSL.
     * 
     * @param extensions
     */
    private void enableSSLSupport(final ConfigurationExtensions extensions) {
        final String trustStore = extensions.get("ssl-trustStore");
        final String trustStorePassword = extensions.get("ssl-trustStorePassword");
        final String keyStore = extensions.get("ssl-keyStore");
        final String keyStorePassword = extensions.get("ssl-keyStorePassword");

        if ((trustStore != null) && (trustStorePassword != null) && (keyStore != null)
                && (keyStorePassword != null)) {
            // TODO
        }
    }
}
